// temu.cc -- P&G VLBA tape drive emulator
// $Id$

// Copyright (C) 1997--2002 Ari Mujunen. (Ari.Mujunen@hut.fi)

// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License.
// See the file 'COPYING' for details.

// $Log$

// Branched off of dammer/dammer.cc:
// Revision 1.35  2002/02/12 12:10:01  amn

#include "registration.h"
RCSID(temu_cc, "$Id$");

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>  // for inet_ntoa()

#include <string.h>  // for traditional 'str*()' calls
#include <math.h>  // for 'pow()'

#include <assert.h>


#define MY_NAME "temu: "


//----------------------------------------------------------------------
#include "cSerialLine.hh"

// xxx
class cTapeDrive {
private:
  cSerialLine sl;

  // Latest incoming RS232 command message.
  char cmdBuf[20];
  int cmdMcb;  // 2 hex digits, the MCB protocol code in last cmdBuf message
  int cmdVal;  // 4 hex digits, the assoc. MCB 16-bit signed value

  // Latest outgoing RS232 reply message.
  char respBuf[100];

  // Tape transport emulation variables.
  char barCode[13];  // xxx: keep all trailing '\0's
  int footage;

public:
  virtual void init(const char *serial_device);
  virtual void cleanup(void);
  virtual ~cTapeDrive();

  // Routines which write replies to serial line.
  virtual void writeString(const char *s);
  virtual void writePrompt(void);
  virtual void greetingMessage(void);
  virtual void writeErrorMessage(void);

  // MCB command handlers.
  virtual void doRCommand(int mcbCode);
  virtual void doWCommand(int mcbCode, int mcbVal);

  // The main processing loop of reading commands from serial line
  // and dispatching them to handlers.
  virtual void readLoop(void);  // the main processing loop
};  // class cTapeDrive


void
cTapeDrive::init(const char *serial_device)
{
  this->sl.open(serial_device);
  // Can later use 'sl.whatPathname()' to access the stored serial device path.
  this->sl.serialSettings(/*baudRate=>*/ 9600,
			  /*numberOfDataBits=>*/ 8,
			  /*numberOfStopBits=>*/ 1,
			  /*whichParity=>*/ eParityNone,
			  /*minBytes=>*/ 1,
			  /*maxTimeout=>*/ 2);  // 0.2 seconds
  // Our timeout strategy is that all reads must return at least one byte,
  // and more consecutive bytes will appear within 0.2 sec.

  // Tape transport emulation variables.
  memset(this->barCode, '\0', sizeof(this->barCode));
  this->footage = 0x1234;  // xxx: debug, should set to 0
}  // init

void
cTapeDrive::cleanup(void)
{
  this->sl.close();
}  // cleanup

cTapeDrive::~cTapeDrive(void)
{
  // xxx: without this virtual destructor, g++ warns:
  // temu.cc:53: warning: `class cTapeDrive' has virtual functions but non-virtual destructor
}  // destructor


void
cTapeDrive::writeString(const char *s)
{
  this->sl.write(s, strlen(s));
}  // writeString

void
cTapeDrive::writePrompt(void)
{
  writeString("MCB>");
}  // writePrompt

void
cTapeDrive::greetingMessage(void)
{
  // writeString("TEMU P&G VLBA Drive MCB RS323 terminal emulator\r\n");
  // Genuine start-up message of a P&G DPU:
  writeString("\r\nV2.0\r\nRS-232 Emulation of MCB active...>\r\n");
  writePrompt();
}  // greetingMessage

void
cTapeDrive::writeErrorMessage(void)
{
  writeString("\r\nValid run string :<R/W> <Address> <NULL/Data>\r\n");
}  // writeErrorMessage


void
cTapeDrive::doRCommand(int mcbCode)
{
  int result;

  // Select what to enquiry and return the result to station unit
  // based on MCB command code, the default result being zero.
  result = 0;
  switch (mcbCode) {
    case 0x30: {  // current footage counter
      result = this->footage;
      break;
    }
    case 0x33: {  // low tape sense, 0=no/1=yes
      result = 0x0000;
      break;
    }
    case 0x34: {  // bar code length in characters
      result = strlen(this->barCode);
      break;
    }
    case 0x35:
    case 0x36:
    case 0x37:
    case 0x38:
    case 0x39:
    case 0x3a: {  // bar code chars (nulls converted to blanks)
      result = (barCode[mcbCode-0x35] ? barCode[mcbCode-0x35]:' ')<<8
	| (barCode[mcbCode-0x35+1] ? barCode[mcbCode-0x35+1]:' ');
      break;
    }
    case 0x71: {  // software revision
      result = 0x0004;  // from P&G DPU
      break;
    }
    case 0x73: {  // general status word
      result = 0x01c0;  // bits 6--8, vac ok, servo ready, 5MHz, 1pps ok
      break;
    }
    case 0x74: {  // error flags
      result = 0x0000;  // no errors
      break;
    }
  }  // switch

  // Send the result (trunc to 16 bits) and a response back.
  result &= 0xffff;
  snprintf(respBuf, sizeof(respBuf), "    Monitor Value = %04X\r\n", result);
  writeString(respBuf);

  // Debug.
  fprintf(stderr, "R %02X --> %04X\n", mcbCode, result);
}  // doRCommand

void
cTapeDrive::doWCommand(int mcbCode, int mcbVal)
{
  // Debug.
  fprintf(stderr, "W %02X %04X\n", mcbCode, mcbVal);

  // xxx: replies nothing
}  // doWCommand

void
cTapeDrive::readLoop(void)
{
  // Flush the incoming buffer, print greeting.
  this->sl.flushPendingInput();
  this->greetingMessage();

  // Process incoming commands forever.
  while (1) {
    // The strategy is to read a single (first) command character
    // which has to be a 'R' for read or 'W' for write.
    // (Sits here blocked/waiting if nothing is available.)
    this->sl.read(cmdBuf, 1);

    if (cmdBuf[0] == 'R') {
      // Read command 'R xx\r\n', receive 5 more bytes from serial.
      this->sl.read(&(cmdBuf[1]), 5);  cmdBuf[6] = '\0';

      if (sscanf(cmdBuf, "R %2x\r\n", &cmdMcb) == 1) {
	this->doRCommand(cmdMcb);
      } else {
	// R command format error.
	this->writeErrorMessage();
	fprintf(stderr, "ignored R command '%s' because of format error\n", cmdBuf);
      }
      this->writePrompt();  // regardless of whether successful or not

    } else if (cmdBuf[0] == 'W') {
      // Write command 'W xx vvvv\r\n', receive 10 more bytes from serial.
      this->sl.read(&(cmdBuf[1]), 10);  cmdBuf[11] = '\0';
      if (sscanf(cmdBuf, "W %2x %4x\r\n", &cmdMcb, &cmdVal) == 2) {
	if (cmdVal > 32768) {
	  // Convert unsigned 16-bit number to signed.
	  cmdVal = cmdVal - 65536;
	}
	this->doWCommand(cmdMcb, cmdVal);
      } else {
	// W command format error.
	this->writeErrorMessage();
	fprintf(stderr, "ignored W command '%s' because of format error\n", cmdBuf);
      }
      this->writePrompt();  // regardless of whether successful or not

    } else {
      // Ignore unknown command character.
      fprintf(stderr, "ignored out-of-sequence character '%c' (0x%02X)\n", cmdBuf[0], (unsigned int)(cmdBuf[0]));
    }
  }  // while
}  // readLoop



//---
char *getUTC(char *p)
{
  struct timeval temptv;
  time_t temptime;
  struct tm *ti;
  static char bu[17];
  int centisec;

  if (p == NULL) {
    p = bu;
  }
  
  assert( gettimeofday(&temptv, NULL) == 0 );
  temptime = (time_t)temptv.tv_sec;
  ti = gmtime(&(temptime));
  centisec = temptv.tv_usec / 10000;
  if (centisec > 99) {
    centisec = 99;
  }
  assert( sprintf(p, "%04d%02d%02d%02d%02d%02d%02d",
          ti->tm_year + 1900,
          ti->tm_mon + 1,
          ti->tm_mday,
          ti->tm_hour,
          ti->tm_min,
          ti->tm_sec,
          centisec
                  // int     tm_wday;        /* day of the week */
                  // int     tm_yday;        /* day in the year */
                  // int     tm_isdst;       /* daylight saving time */
          ) <= ((int)sizeof(bu)-1) );
  return p;
}  // getUTC



// ----------------------------------------------------------------------
int
main (int argc, char **argv)
{
  cTapeDrive td;

  int option_char;

  // Startup version message.
  fprintf (stderr, MY_NAME "$Id$\n");

  // Configuration file / command line option defaults.
  // Command line wins, default conf file is next,
  // and these hard-coded values are the last resort.
  int vsi_port_number = 5555;  // xxx: check the real Mark5A VSI-S port
  char *my_serial_device = strdup("/dev/ttyS0");
  assert (my_serial_device != NULL);

  // Process default conf file.
  // xxx process_conf(/*file=>*/ "/etc/temu.conf" );

  // Invokes member function `int operator ()(void);'
  while ((option_char = getopt (argc, argv, "c:d:p:")) != EOF)
    switch (option_char) {
       case 'c': {
         assert( optarg != NULL );
	 // xxx process_conf(/*file=>*/getopt.optarg);
	 break;
       }
       case 'd': {
	 free(my_serial_device);
	 assert ((my_serial_device = strdup(optarg)) != NULL);
         break;
       }
       case 'p': {
	 vsi_port_number = atoi (optarg);
	 break;
       }
       case ':':  // missing parameter after a option letter
       case '?': {
	 fprintf (stderr,
		  "usage: %s [-d<serial device>] [-p<port>]\n", argv[0]);
	 exit(EXIT_FAILURE);
       }
    }  // switch

  // Prime the tape drive emulator to use the serial line device provided.
  td.init(my_serial_device);

  // Start listening to incoming serial bytes and react to them.
  td.readLoop();

  td.cleanup();  // xxx ever executed?
}  // main
