/* reavsib -- Read VLBI VSIBrute input board. */

/*
 * $Log: reavsib.c,v $
 * Revision 1.5  2002/03/28 09:03:47  amn
 * Preparing for multi-day sustained 640Mbit/s tests.
 *
 * Revision 1.4  2002/03/25 15:18:51  amn
 * First chained DMA ring buffer.
 *
 * Revision 1.3  2002/03/21 11:23:20  amn
 * Added mode=, giga=, 1pps= and new-style command word bit assignments.
 *
 * Revision 1.2  2002/02/27 14:33:28  amn
 * Changed Log line to be on next line than the comment start characters.
 *
 * Revision 1.1  2002/02/27 14:24:38  amn
 * Initial version.
 *
 */

/* Modified from: */
/* reax13 -- Read AMS HRDL X13 input board. */
/* Modified from: */
/* dd -- convert a file while copying it.
   Copyright (C) 85, 90, 91, 95, 1996 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* Written by Paul Rubin, David MacKenzie, and Stuart Kemp. */
/* Hacked to bare minimum for Debian boot floppies by Giuliano Procida */

/* Options:

   Numbers can be followed by a multiplier:
   b=512, c=1, k=1024, w=2, xm=number m

   bs=BYTES                     Read and Write BYTES bytes at a time.
   count=BLOCKS                 Copy only BLOCKS input blocks. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>
#include <sched.h>
#include <sys/ioctl.h>

#include "vsib_ioctl.h"

#define equal(p, q) (strcmp ((p),(q)) == 0)

/* The name this program was run with. */
char *program_name;

/* The "original" name of the program. */
#define MYNAME "reavsib"

#define DO_SLEEPS 1
#define SMALL_SLEEP (1000L)

/* How many 'invdata' reads required for on-the-fly data pol inversion? */
#define NUM_OF_CONSECUTIVE_INVDATAS_NEEDED (5)


static void
usage(const char *p) {
  fprintf(stderr, p);
}

const char      reavsib_usage[] = MYNAME " [OPTION]\n\
\n\
Read from VSIB stdin to stdout according to the options.\n\
\n\
  bs=BYTES        read and write BYTES bytes at a time\n\
                    (default 128k)\n\
  skip=BYTES      skip BYTES samples between ones taken\n\
                    (default 0, nothing skipped)\n\
  mode=n          set VSIB board capture mode\n\
                    (default 0 = all32bits)\n\
  count=BLOCKS    copy only BLOCKS input blocks\n\
                    (default -1, unlimited)\n\
  timeout=READS   start complaining after READS number of empty reads\n\
                    (default -1, unlimited)\n\
  1pps=1          set 'ffff ffff' markers for 1pps TOST data\n\
                    (default 0 = no markers, real data)\n\
  giga=1          delay out-VSI 1pps by 1 clock for gigabit\n\
                    (default 0 = 1pps outputted with TOST sample)\n\
\n\
BYTES may be suffixed: by xM for multiplication by M, by c for x1,\n\
by w for x2, by b for x512, by k for x1024.\n";

static int parse_integer (char *str);
static int copy (char *buf);

/* Default number of bytes in which atomic reads and writes are done. */
static long blocksize = 128*1024;  /* the maximum of X13 DMA buffer */

/* Default number of samples to skip between ones taken. */
static long skip_samples = 0;

/* VSIB board capture mode, default 0 == all32bits. */
static int vsib_mode = 0;

/* VSIB board gigabit mode on? */
static int vsib_giga = 0;

/* VSIB board insert 1pps 'ffff ffff' markers mode on? */
static int vsib_1pps = 0;

/* Copy only this many records.  <0 means no limit. */
static int max_records = -1;

/* How many zero-length 'read()' calls will cause error messages?
   (<0 means no limit.) */
static int timeout_number_of_zero_reads = -1;

/* Number of partial blocks written. */
static unsigned w_partial = 0;

/* Number of full blocks written. */
static unsigned w_full = 0;

/* Number of partial blocks read. */
static unsigned r_partial = 0;

/* Largest partial block read. */
static unsigned r_largest_partial = 0;

/* Largest number of consecutive full blocks read. */
static unsigned r_largest_consecutive_full = 0;

/* Number of full blocks read. */
static unsigned r_full = 0;


/* A protected (error-checking) ioctl() for VSIB driver. */
static void
vsib_ioctl(
  unsigned int mode,
  unsigned long arg
) {
  if (ioctl(STDIN_FILENO,
            mode,
            arg)
  ) {
    char err[255];
    
    sprintf(err, MYNAME ": ioctl(STDIN_FILENO, 0x%04x,...)", mode);
    perror(err);
    fprintf(stderr, MYNAME ": standard input is not an VSIB board\n");
    exit(EXIT_FAILURE);
  }
}  /* vsib_ioctl */


extern int
main (int argc, char **argv)
{
  int exit_status;
  unsigned char *buf;           /* Buffer. */
  int i, n;
  
  program_name = argv[0];

  fprintf (stderr, MYNAME ": $Revision: 1.5 $ $Date: 2002/03/28 09:03:47 $\n");

  /* Decode arguments. */
  for (i = 1; i < argc; i++)
    {
      char *name, *val;
      
      name = argv[i];
      val = strchr (name, '=');
      if (val == 0)
        {
          fprintf (stderr, "%s: unrecognized option `%s'\n", program_name, name);
          usage (reavsib_usage);
          return 1;
        }
      *val++ = '\0';
      
      n = parse_integer (val);
      if (n < 0)
        { 
          fprintf (stderr, "%s: invalid number: `%s'\n", program_name, val);
          return 1;
        }
      
      if (equal (name, "bs"))
        blocksize = n;
      else if (equal (name, "skip"))
        skip_samples = n;
      else if (equal (name, "mode"))
        vsib_mode = n;
      else if (equal (name, "giga"))
        vsib_giga = n;
      else if (equal (name, "1pps"))
        vsib_1pps = n;
      else if (equal (name, "count"))
        max_records = n;
      else if (equal (name, "timeout"))
        timeout_number_of_zero_reads = n;
      else
        {
          fprintf (stderr, "%s: unrecognized option `%s=%s'\n", program_name, name, val);
          usage (reavsib_usage);
          return 1;
        }
    }
  
#if 0
  /* Boost us a little higher than general Netscape processes et al... */
  /***nice(-10); doesn't affect anything at all */

  /* Set POSIX the lowest "real-time" priority */
  /* which wins all standard SCHED_OTHER processes. */
  {
    struct sched_param s;

    s.sched_priority = sched_get_priority_max(SCHED_FIFO);
    if (sched_setscheduler(/*pid=>*/0, SCHED_FIFO, &s)) {
      perror(MYNAME ": sched_setscheduler(0, SCHED_FIFO, ...)");
      /***return (EXIT_FAILURE);***/
    }
  }
  /* xxx: even with SCHED_FIFO, usleep()s and blocking read()s
     are a problem; 300ms or so can elapse before waking up... */
#endif

  /* Allocate buffer space and do the copy. */
  buf = (unsigned char *) malloc (blocksize);
  if ( buf == 0 ) {
    fprintf (stderr, "%s: Memory exhausted\n", program_name);
    exit_status = 1;
  } else {
    exit_status = copy (buf);
    free (buf);
  }
  
  fprintf (stderr, "%u+%u records in, largest partial %u bytes, largest consecutive full block row=%u\n",
           r_full,
           r_partial,
           r_largest_partial,
           r_largest_consecutive_full);
  fprintf (stderr, "%u+%u records out\n", w_full, w_partial);
  {
    unsigned long rr;

    vsib_ioctl(VSIB_GET_DMA_RETRIES, (unsigned long)(&rr));
    fprintf (stderr, "%lu DMA chip counter read retries\n", rr);
  }
  
  return exit_status;
}

/* This routine inits X13 board, making sure its invclock and
   invdata indicators say "normal". */
static unsigned long /* ioctl arg */ current_mode_reg;
static void
start_vsib_with_inv_fixups(void)
{
  unsigned char status_byte;

  /* Overall strategy: stop, re-init DMA, start. */

  /* After opening the board is stopped with mode == 0x00,
     but just to make sure, stop it into normal non-div mode... */
  current_mode_reg = VSIB_MODE_STOP;
  vsib_ioctl(VSIB_SET_MODE, current_mode_reg);

  /* The board must have time to detect "non-inv" data. */
  usleep(SMALL_SLEEP);

  /* Query the board, invdata, invclock... */
  vsib_ioctl(VSIB_GET_STATUS, (unsigned long)(&status_byte));

#if 0
  if (status_byte & VSIB_STATUS_INVDATA) {
    fprintf(stderr, MYNAME ": detected inverted data polarity, fixing it\n");
    current_mode_reg |= VSIB_MODE_INVDATA;
  } else {
    fprintf(stderr, MYNAME ": detected normal data polarity\n");
  }
  vsib_ioctl(VSIB_SET_MODE, current_mode_reg);
#endif

  /* Check the results. */
  {
    int i;
    int invclock = 0;
    int invdata = 0;

    for (i=0; i < 10; i++) {
      vsib_ioctl(VSIB_GET_STATUS, (unsigned long)(&status_byte));
      if (status_byte & VSIB_STATUS_INVDATA) {
	invdata++;
      }
    }
    fprintf(stderr, MYNAME ": after fixups, detected %d invclocks and %d invdatas.\n", invclock, invdata);
  }
  
  /* Re-init DMA. */
  vsib_ioctl(VSIB_RESET_DMA, 0);

  /* Start the board.  (After opening it is still stopped.) */
  current_mode_reg
    |= (VSIB_MODE_MODE(vsib_mode)
        | VSIB_MODE_RUN
        | (vsib_giga ? VSIB_MODE_GIGABIT : 0)
        | (vsib_1pps ? VSIB_MODE_EMBED_1PPS_MARKERS : 0)
        | (skip_samples & 0x0000ffff));
  vsib_ioctl(VSIB_SET_MODE, current_mode_reg);
  /* xxx: could perhaps wait for VSIB_STATUS_RXSTATE == 1? */
}  /* start_vsib_with_inv_fixups */


/* The main loop.  */

static int
copy (char *buf)
{
  unsigned char status_byte;
  int nread;                    /* Bytes read in the current block. */
  int nwrite;                   /* Bytes written in the current block. */
  int consecutive_full = 0;  
  int consecutive_zero_reads = 0;
  int consecutive_invdatas = 0;
  int cons_was;
  int receiving = 0;  /* state variable */

  start_vsib_with_inv_fixups();

  /* Do the "copying". */
  consecutive_zero_reads = 0;
  consecutive_invdatas = 0;
  receiving = 0;
  while (1)
    {
      if (max_records >= 0 && r_partial + r_full >= max_records)
        break;

      /* Timeout check: we haven't got anything in
	 'consecutive_zero_reads' number of 'read()' calls. */
      if (timeout_number_of_zero_reads > 1) {
	if (consecutive_zero_reads > timeout_number_of_zero_reads) {
	  if (receiving) {
	    fprintf(stderr, MYNAME ": has not received anything in %d reads, reinit DMA and wait...\n",
		    consecutive_zero_reads);
	    vsib_ioctl(VSIB_RESET_DMA, 0);
	  }
	  receiving = 0;
	}
      }

      /* xxx: Before reading, could get VSIB_STATUS_DMAOVERRUN bit
              (if it worked on X13 board). */

      /* But instead we keep our eye on VSIB_STATUS_INVDATA bit:
	 if it stays on for 10 consecutive reads, we flip the
	 corresponding mode bit. */
      vsib_ioctl(VSIB_GET_STATUS, (unsigned long)(&status_byte));
      if (status_byte & VSIB_STATUS_INVDATA) {
	consecutive_invdatas++;
	fprintf(stderr, MYNAME ": %d read(s) with 'invdata' bit on\n",
		consecutive_invdatas);
	if (consecutive_invdatas >= NUM_OF_CONSECUTIVE_INVDATAS_NEEDED) {
	  fprintf(stderr, MYNAME ": detected inverted data (%d reads), flipping data polarity...\n",
		  consecutive_invdatas);
	  {
#if 0
	    current_mode_reg ^= (VSIB_MODE_INVDATA);
	    vsib_ioctl(VSIB_SET_MODE, current_mode_reg);
#endif
	  }
	  consecutive_invdatas = 0;
	}
      } else {
	consecutive_invdatas = 0;
      }

      nread = read (STDIN_FILENO, buf, blocksize);

      /* ...and if it occurred, could dump this read buffer into a file. */
      /* xxx: also could VSIB_GET_DMA_RETRIES and print it occasionally */

      /* X13 board never ends... */
#if 0
      if (nread == 0)
        break;          /* EOF.  */
#endif
      /* If we got nothing we have to wait a little for X13 */
      /* DMA buffer to fill up. */
      if (nread == 0) {
#if DO_SLEEPS
        usleep(SMALL_SLEEP);
#endif
	consecutive_zero_reads++;
        continue;
      }

      if (nread < 0)
        return -2;

      /* Now we got something, nread > 0. */
      cons_was = consecutive_zero_reads;  /* from here... */
      consecutive_zero_reads = 0;
      if (!receiving) {
	int i;
	int invclock = 0;
	int invdata = 0;

	/* Now we are back in business, but has invclock/invdata
	   remained the same? */
	fprintf(stderr, MYNAME ": did not receive anything in %d reads, now continuing...\n",
		cons_was);  /* ...to here */

	/* The board must have time to detect "non-inv" data. */
	usleep(SMALL_SLEEP);
	/* xxx: doesn't help, occasionally X13 says "it is inverted" */

	for (i=0; i < 10; i++) {
	  vsib_ioctl(VSIB_GET_STATUS, (unsigned long)(&status_byte));
	  if (status_byte & VSIB_STATUS_INVDATA) {
	    invdata++;
	  }
	}  /* for */
	fprintf(stderr, MYNAME ": after restart, detected %d invclocks and %d invdatas.\n", invclock, invdata);
	if ((invclock > 2) || (invdata > 2)) {
	  fprintf(stderr, MYNAME ": too many invclocks/invdatas, junking %d bytes and re-fixing...\n", nread);

	  start_vsib_with_inv_fixups();

	  consecutive_zero_reads = 0;
	  receiving = 0;
	  continue;
	}  /* if invclock/data not stable after restart */
      }
      receiving = 1;

      if (nread < blocksize) {
        r_partial++;

        /* A partial block terminates a row of full blocks. */
        consecutive_full = 0;

        /* Remember largest partial block. */
        if (nread > r_largest_partial) {
          r_largest_partial = nread;
        }
      } else {
        r_full++;
        consecutive_full++;

        /* Remember the largest row of full blocks. */
        if (consecutive_full > r_largest_consecutive_full) {
          r_largest_consecutive_full = consecutive_full;
        }
      }

      nwrite = write (STDOUT_FILENO, buf, nread);

      if (nwrite < 0)
        return -1;

      if (nwrite == blocksize)
        w_full++;
      else
        w_partial++;

      /* If we got less than what we asked for we emptied the X13 */
      /* DMA buffer and thus we can sleep a little before next read. */
#if DO_SLEEPS
      /* xxx: Was '<=' in AMS code and thus slept always...! */
      if (nread < blocksize) {
        usleep(SMALL_SLEEP);
      }
#endif

    }
  return 0;
}

/* Return the value of STR, interpreted as a non-negative decimal integer,
   optionally multiplied by various values.
   Return -1 if STR does not represent a number in this format. */

static int
parse_integer (char *str)
{
  register int n = 0;
  register int temp;
  char *p = str;

  n = strtoul (p,&p,10);

loop:
  switch (*p++)
    {
    case '\0':
      return n;
    case 'b':
      n *= 512;
      goto loop;
    case 'c':
      goto loop;
    case 'k':
      n *= 1024;
      goto loop;
    case 'w':
      n *= 2;
      goto loop;
    case 'x':
      temp = parse_integer (p);
      if (temp == -1)
        return -1;
      n *= temp;
      break;
    default:
      return -1;
    }
  return n;
}
