#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <sys/types.h>  /* open() */
#include <sys/stat.h>  /* open() */
#include <fcntl.h>  /* open() */

#include <sys/time.h>  /* gettimeofday() */
#include <unistd.h>  /* gettimeofday(), usleep() */

#include <string.h>  /* strstr() */


typedef unsigned char tByte, *ptByte;

/* Data words, 32/16/8 bits. */
typedef unsigned int tD, *ptD;


void
crc16(ptD cr, ptD d, int datawords) 
{
  int i;

  /* 'cr' CRC register is assumed to be initialized */
  /* with 16 data words. */
  /* (Init to all zero; or use old regs to continue. */
  for (i=0; i<datawords; i++) {
    tD n0  = *d++ ^ cr[15];  /* x^16 + */
    tD n15 = n0 ^ cr[14];  /* x^15 + */
    tD n2  = n0 ^ cr[1];   /* x^2 + 1 */

    cr[15] = n15;
    cr[14] = cr[13];
    cr[13] = cr[12];
    cr[12] = cr[11];
    cr[11] = cr[10];
    cr[10] = cr[ 9];
    cr[ 9] = cr[ 8];
    cr[ 8] = cr[ 7];
    cr[ 7] = cr[ 6];
    cr[ 6] = cr[ 5];
    cr[ 5] = cr[ 4];
    cr[ 4] = cr[ 3];
    cr[ 3] = cr[ 2];
    cr[ 2] = n2;
    cr[ 1] = cr[ 0];
    cr[ 0] = n0;
  }  /* for every data word (with multiple bits) in input data stream */
  /* The result is in 'cr[]', MSB in last data word. */
}  /* crc16 */

void
crc12(ptD cr, ptD d, int datawords) 
{
  int i;

  /* 'cr' CRC register is assumed to be initialized */
  /* with 12 data words. */
  /* (Init to all zero; or use old regs to continue. */
  for (i=0; i<datawords; i++) {
    tD n0  = *d++ ^ cr[11];  /* x^12 + */
    tD n11 = n0 ^ cr[10];  /* x^11 + */
    tD n3  = n0 ^ cr[2];   /* x^3 + */
    tD n2  = n0 ^ cr[1];   /* x^2 + */
    tD n1  = n0 ^ cr[0];   /* x + 1 */

    cr[11] = n11;
    cr[10] = cr[ 9];
    cr[ 9] = cr[ 8];
    cr[ 8] = cr[ 7];
    cr[ 7] = cr[ 6];
    cr[ 6] = cr[ 5];
    cr[ 5] = cr[ 4];
    cr[ 4] = cr[ 3];
    cr[ 3] = n3;
    cr[ 2] = n2;
    cr[ 1] = n1;
    cr[ 0] = n0;
  }  /* for every data word (with multiple bits) in input data stream */
  /* The result is in 'cr[]', MSB in last data word. */
}  /* crc16 */

/* CRC function poiter type def. */
typedef void (*pCRCfunc)(ptD cr, ptD d, int datawords);



/* Transpose template bytes bit at a time (starting from MSB) */
/* into 'target' data words, either a target bit stream at a time */
/* or all bits together.  Returns ptr of next unprocessed vacant data word. */
ptD
fmt(ptD target, ptByte from, int bytes,
    int whichbit, int whichparity, ptD crcBlock, pCRCfunc pcrcf
)
{
  tD mask;
  int i, j;

  if (whichbit < 0) {
    /* Change all parallel bit streams of data. */
    mask = 0xffffffff;
  } else {
    /* Change only a single bit stream. */
    assert( whichbit < sizeof(tD)*8 );
    mask = 1 << whichbit;
  }
  for (i=0; i<bytes; i++) {
    tD parity;
    tByte f;

    /* <0 even (for sync word), ==0 none, >0 odd (for all other data). */
    if (whichparity > 0) {
      parity = mask;
    } else {
      parity = 0;
    }

    f = from[i];
    /* Put every bit of source template into a new bit stream target word. */
    for (j=0; j<=7; j++) {
      tD d;

      d = ( (f & 0x80) ? mask : 0 );
      parity ^= d;
      if (whichbit < 1) {  /* all==-1 or first==0, replace/zero all */
	*target = d;
      } else {
	/* Single-bits 1--31, just add a single bit. */
	*target |= d;
      }
      if (crcBlock) {  /* used only with -1==all or 31==last bit */
	(*pcrcf)(crcBlock, target, 1);
      }
      target++;
      f <<= 1;  /* next bit */
    }  /* for every bit in template from byte */

    /* For both even and odd parity, add parity bit data word. */
    if (whichparity != 0) {
      if (whichbit < 1) {  /* all==-1 or first==0, replace/zero all */
	*target++ = (parity & mask);
      } else {
	/* Single-bits 1--31, just add a single bit. */
	*target++ |= (parity & mask);
      }
    }
  }  /* for every target bit */
#if 0
  printf("fmt: %08X after inserting %d bytes\n", (unsigned int)target, bytes);
#endif
  return (target);
}  /* fmt */


/* Mark IV frame header in byte chunks, without parity bits. */
typedef struct sRawMk4Header {
  /* 64-bit/8-byte aux field */
  tByte auxhs1[2];
  tByte auxhs2[2];
  tByte auxtrk;
  tByte auxsrc;
  tByte auxsts;
  tByte auxsid;
  /* 4-byte sync word */
  tByte sync[4];
  /* 64-bit/8-byte time code (52 bits) plus CRC (12 bits) field */
  tByte timec[5];  /* to full seconds */
  tByte timef[3];  /* sub-second plus CRC-12 */
} tRawMk4Header;


ptD
mk4fmtc(ptD target, ptD crcBlock, int useParity)
{
  tRawMk4Header f;
  int trk;

  /* Init CRC generator to all zeroes. */
  memset(crcBlock, 0, 12*sizeof(tD));

  /* Dummy headstack positions. */
  f.auxhs1[0] = 0x55; f.auxhs1[1] = 0x55;
  f.auxhs2[0] = 0x55; f.auxhs2[1] = 0x55;
  target = fmt(target, (ptByte)(&(f.auxhs1)), 4, -1,
	       useParity?/*odd*/1:0, crcBlock, &(crc12));

  /* Run track information separately for all bits in data word. */
  for (trk=0; trk<=30; trk++) {
    f.auxtrk = trk;
    f.auxsrc = trk;
    (void)fmt(target, (ptByte)(&(f.auxtrk)), 2, trk,
	      useParity?/*odd*/1:0, NULL, NULL);
  }  /* for each track */
  /* Calculate CRC only when the last bit/track 31 is in place. */
  f.auxtrk = trk;
  f.auxsrc = trk;
  target = fmt(target, (ptByte)(&(f.auxtrk)), 2, trk,
	       useParity?/*odd*/1:0, crcBlock, &(crc12));
  f.auxsts = 0x00;  /* no formatter errors */
  f.auxsid = 0x12;  /* system ID */
  target = fmt(target, (ptByte)(&(f.auxtrk)), 2, -1,
	       useParity?/*odd*/1:0, crcBlock, &(crc12));

  f.sync[0] = 0xff; f.sync[1] = 0xff; f.sync[2] = 0xff; f.sync[3] = 0xff;
  target = fmt(target, (ptByte)(&(f.sync)), 4, -1,
	       useParity?/*even*/-1:0, crcBlock, &(crc12));

  f.timec[0] = 0x21; f.timec[1] = 0x23;  /* YDDD */
  f.timec[2] = 0x19; f.timec[3] = 0x27;  /* HHMM */
  f.timec[4] = 0x15;  /* SS */
  target = fmt(target, (ptByte)(&(f.timec)), 5, -1,
	       useParity?/*odd*/1:0, crcBlock, &(crc12));
  /* 'crcBlock' contains "seed" of CRC state this far. */
  return (target);
}  /* mk4fmtc */

ptD
mk4fmtf(ptD target, ptD crcBlock, int useParity)
{
  tD parity;
  tD mask = 0xffffffff;
  tByte f;
  int j;

  /* First byte, odd parity. */
  parity = mask;
  f = 0x12;  /* 0.1 and 0.01 seconds */
  for (j=0; j<=7; j++) {
    tD d;

    d = ( (f & 0x80) ? mask : 0 );
    parity ^= d;
    *target = d;
    crc12(crcBlock, target, 1);
    target++;
    f <<= 1;  /* next bit */
  }  /* for every bit in template from byte */
  if (useParity) {
    *target++ |= parity;
  }

  /* Second byte, odd parity. */
  parity = mask;
  f = 0x30;  /* specially-encoded 0.001 seconds */
  for (j=0; j<=3; j++) {
    tD d;

    d = ( (f & 0x80) ? mask : 0 );
    parity ^= d;
    *target = d;
    crc12(crcBlock, target, 1);
    target++;
    f <<= 1;  /* next bit */
  }  /* for four MSB bits in template from byte */
  /* Add four MSB bits from CRC generator. */
  *target++ = crcBlock[11];  parity ^= crcBlock[11];
  *target++ = crcBlock[10];  parity ^= crcBlock[10];
  *target++ = crcBlock[9];  parity ^= crcBlock[9];
  *target++ = crcBlock[8];  parity ^= crcBlock[8];
  if (useParity) {
    *target++ = parity;
  }

  /* Third byte, rest of CRC with odd parity. */
  parity = mask;
  for (j=7; j>=0; j--) {
    *target++ = crcBlock[j];  parity ^= crcBlock[j];
  }
  if (useParity) {
    *target++ = parity;
  }
  return (target);
}  /* mk4fmtf */



/* VLBA frame header in byte chunks, without parity bits. */
typedef struct sRawVLBAHeader {
  /* 4-byte sync word */
  tByte sync[4];
  /* 64-bit/8-byte time code (48 bits) plus CRC (12 bits) field */
  tByte timec[4];  /* MJD to full seconds */
  tByte timef[2];  /* sub-second */
  tByte crc16[2];  /* plus CRC-16 (2 bytes) */
  /* Here in the middle 22500 bits of non-replacement data + parity. */
  /* 64-bit/8-byte aux field */
  tByte auxsid;
  tByte auxhs[3];  /* recorder#, headstack#, hs position */
  tByte auxdar[3];  /* bit-reversed TG BS CD */
  tByte auxunused;  /* 0x00 */
} tRawVLBAHeader;


ptD
vlbafmtc(ptD target, ptD crcBlock, int useParity)
{
  tRawVLBAHeader f;

  /* Init CRC generator to all zeroes. */
  memset(crcBlock, 0, 16*sizeof(tD));

  f.sync[0] = 0xff; f.sync[1] = 0xff; f.sync[2] = 0xff; f.sync[3] = 0xff;
  target = fmt(target, (ptByte)(&(f.sync)), 4, -1,
	       useParity?/*even*/-1:0, crcBlock, &(crc16));

  f.timec[0] = 0x21; f.timec[1] = 0x23;  /* JJJS */
  f.timec[2] = 0x19; f.timec[3] = 0x27;  /* SSSS */
  target = fmt(target, (ptByte)(&(f.timec)), 4, -1,
	       useParity?/*odd*/1:0, crcBlock, &(crc16));
  /* 'crcBlock' contains "seed" of CRC state this far. */
  return (target);
}  /* vlbafmtc */

ptD
vlbafmtf(ptD target, ptD crcBlock, int useParity)
{
  tByte f[2];  /* 0.ssss seconds in BCD */
  tD parity;
  tD mask = 0xffffffff;
  int j;

  /* First two bytes, odd parity. */
  f[0] = 0x21; f[1] = 0x23;  /* .ssss */
  target = fmt(target, f, 2, -1,
	       useParity?/*odd*/1:0, crcBlock, &(crc16));

  /* 3rd byte, odd parity. */
  parity = mask;
  for (j=15; j>=8; j--) {
    *target++ = crcBlock[j];  parity ^= crcBlock[j];
  }
  if (useParity) {
    *target++ = parity;
  }

  /* 4th byte, rest of CRC with odd parity. */
  parity = mask;
  for (j=7; j>=0; j--) {
    *target++ = crcBlock[j];  parity ^= crcBlock[j];
  }
  if (useParity) {
    *target++ = parity;
  }
  return (target);
}  /* vlbafmtf */

ptD
vlbafmta(ptD target, int useParity)
{
  tRawVLBAHeader f;
  int trk;

  f.auxsid = 0x12;  /* system ID */
  /* Recorder#, headstack#, hs position. */
  f.auxhs[0] = 0x00; f.auxhs[1] = 0x00; f.auxhs[2] = 0x00;
  target = fmt(target, (ptByte)(&(f.auxsid)), 4, -1,
	       useParity?/*odd*/1:0, /*crcBlock=>*/NULL, NULL);

  /* Run track information separately for all bits in data word. */
  for (trk=0; trk<=30; trk++) {
    f.auxdar[0] = trk;  /* BR TG */
    f.auxdar[1] = trk;  /* BR BS */
    f.auxdar[2] = trk;  /* BR CD */
    (void)fmt(target, (ptByte)(&(f.auxdar)), 3, trk,
	      useParity?/*odd*/1:0, NULL, NULL);
  }  /* for each track */
  f.auxdar[0] = trk;  /* BR TG */
  f.auxdar[1] = trk;  /* BR BS */
  f.auxdar[2] = trk;  /* BR CD */
  target = fmt(target, (ptByte)(&(f.auxdar)), 3, trk,
	       useParity?/*odd*/1:0, /*crcBlock=>*/NULL, NULL);

  f.auxunused = 0x00;  /* always zero */
  target = fmt(target, (ptByte)(&(f.auxunused)), 1, -1,
	       useParity?/*odd*/1:0, /*crcBlock=>*/NULL, NULL);
  return (target);
}  /* vlbafmta */



int
main(int argc, char **argv)
{
  int useParity;
  size_t framesize;
  ptD frame;
  ptD crcBl;
  ptD t;
  int fid;

  useParity = 1;
#if 0
  framesize = 2500*(useParity?9:8);
  assert( (frame = calloc(framesize, sizeof(tD))) != NULL );
  assert( (crcBl = calloc(12, sizeof(tD))) != NULL );
  printf("Start address = %08X\n", (unsigned int)frame);
  t = mk4fmtc(frame, crcBl, useParity);
  printf("TC  address   = %08X, bits = %d\n", (unsigned int)t, t-frame);
  t = mk4fmtf(t, crcBl, useParity);
  printf("End address   = %08X, bits = %d\n", (unsigned int)t, t-frame);

  assert( (fid = creat("mk4frame", S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) != -1 );
  assert( (write(fid, frame, framesize*sizeof(tD)) == framesize*sizeof(tD)) );
  assert( (close(fid) != -1) );
#endif

  framesize = (2500 + 20)*(useParity?9:8);
  assert( (frame = calloc(framesize, sizeof(tD))) != NULL );
  assert( (crcBl = calloc(16, sizeof(tD))) != NULL );
  printf("Start address = %08X\n", (unsigned int)frame);
  t = vlbafmtc(frame, crcBl, useParity);
  printf("TC  address   = %08X, bits = %d\n", (unsigned int)t, t-frame);
  t = vlbafmtf(t, crcBl, useParity);
  printf("End address   = %08X, bits = %d\n", (unsigned int)t, t-frame);
  t = &(frame[(2500 + 12)*(useParity?9:8)]);
  printf("Aux start address = %08X\n", (unsigned int)t);
  t = vlbafmta(t, useParity);
  printf("Aux end address   = %08X, bits = %d\n", (unsigned int)t, t-&(frame[(2500 + 12)*(useParity?9:8)]));

  assert( (fid = creat("vlbaframe", S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) != -1 );
  assert( (write(fid, frame, framesize*sizeof(tD))) == (framesize*sizeof(tD)) );
  assert( (close(fid) != -1) );

  exit (EXIT_SUCCESS);
}  /* main */
