/* vsibcli.cc  --  Command line interface (via TCP/IP) for vsibcon. */

// $Log$


#include "registration.h"
RCSID(vsibcli_cc, "$Id$");

// Configuration/debugging defines.
// none yet

#define MY_NAME "vsibcli: "


#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <string.h>

// For shared memory.
#include <sys/ipc.h>
#include <sys/shm.h>


// Socket server common code.
// (Consists of classes 'cMessageProcessor' and 'cSocketServer'.)
#include "ss.cc"


/*------------------------------------------------------------*/
// The common shared struct
// across vsibcon+vsibcli.
//
#include "sh.h"

// A pointer to it, for shared memory references in Linux with 'sh->'.
ptSh sh = NULL;


// --------------------------------------------------
// All-purpose routines for reading numbers and alike.

typedef enum {
  eMHz
} tFreqType;

int
getFreq(
  const char *word,
  tFreqType deflt,  // MHz or something else
  double *result
) {
  double temp;

  if (sscanf(word, "%lg", &temp) == 1) {
    // Conversion successful.
    // xxx: For now, we'll just allow everything.
    *result = temp;
    return 0;
  } else {
    // Conversion failed, leave the result alone/unchanged.
    return 1;
  }
}  // getFreq


// Timestamp <--> ASCII formatting.
#include "tstamp.c"


/*------------------------------------------------------------*/
// Vsibcon's command line message processor is based on the generic
// TCP/IP socket server 'cMessageProcessor'.
class cMPVsib : public cMessageProcessor {
private:
  // Background/logging.
  int nextChannel;
public:
  virtual void init(void);
  virtual void cleanup(void);
  virtual ~cMPVsib();
  virtual int process_message_buffer(int filedes, char *buffer, int nbytes);
  virtual void doBackground(void);
};  // class cMPVsib


void
cMPVsib::init(void)
{
  this->nextChannel = 0;
}  // init

void
cMPVsib::cleanup(void)
{
}  // cleanup

cMPVsib::~cMPVsib(void)
{
  // xxx: ???
}  // destructor

void
sureWrite(int fd, const void *buf, size_t count)
{
  size_t wr;
  
  wr = write(fd, buf, count);
  if (wr != count) {
    fprintf(stderr, MY_NAME "sureWrite: wrote %d bytes, should have written %d ('%*s')\n", wr, count, count, (char *)buf);
  }
}  // sureWrite


int
cMPVsib::process_message_buffer(int filedes,
                               char *buffer,
                               int nbytes)
{
  char *words[NWORDS];
  const char delims[] = " \t,=\n\r";
  int number_of_words = 0;
  char *tok;
  char *running = NULL;
  int retCode;
  char replyBuffer[256];
  int rbytes = 0;

  if (nbytes < 1) {
    return 0;  // null request, processed okay
  }

  buffer[nbytes-1] = '\0';  // ensure that 'buffer' ends as a string
  // (this is written over a line termination character)

#if 1
  // Log the incoming command line.
  fprintf(stdout, "%s\t%s\n", getUTC(NULL), buffer);
  fflush(stdout);
#endif

  // Separate command line into words / tokens.
  // Make a copy since 'strsep()' modifies its buffer (puts in '\0' chars).
  assert( (running = strdup(buffer)) != NULL );
  while ( (tok = strsep(&running, delims)) ) {
    if (number_of_words < NWORDS) {
      if (*tok != '\0') {  // skip null arguments
        words[number_of_words] = tok;
        number_of_words++;
      }
    }
  }  // while tokens available

#if 0
  fprintf(stderr, "number_of_words = %d\n", number_of_words);
  {
    int i;

    for (i=0; i<number_of_words; i++) {
      fprintf(stderr, "words[%d] = '%s'\n", i, words[i]);
    }
  }
#endif

  // xxx: Should actually accumulate bytes until a '\n' is found,
  //      and then process each line as a message.
  //      (Note that WinNT 4.0 'telnet' sends each character immediately
  //       and thus using it requires this change.)

  retCode = 0;  // the default return code

  if (number_of_words < 1) {
    retCode = 0;  // null request, processed okay
    goto lExit;
  }

  // Check the first word, that is, the 'vsibcon' command word.

  // STANDBY
  if ((strcmp(words[0], "standby")==0) || (strcmp(words[0], "STANDBY")==0) ) {
    // Reply to network.
    sureWrite(filedes, "OK\n", 3);

    // Signal 'vsibcon' by incrementing the update flag counter.
    sh->vsibcliMode = ACM_standby;
    sh->vsibcliUpdates++;
    goto lExit;
  }  // if "standby"

  // VAXTRACK
  if ((strcmp(words[0], "vaxtrack")==0) || (strcmp(words[0], "VAXTRACK")==0) ) {
    // Reply to network.
    sureWrite(filedes, "OK\n", 3);

    // Signal 'vsibcon' by incrementing the update flag counter.
    sh->vsibcliMode = ACM_VAX_track;
    sh->vsibcliUpdates++;
    goto lExit;
  }  // if "vaxtrack"

  // AZEL
  if ((strcmp(words[0], "azel")==0) || (strcmp(words[0], "AZEL")==0) ) {

    // xxx: "service", az=38.5, el=0.5?

    if (number_of_words < 2) {
      // Report back the current fixed position settings.
      // xxx: or should we return the "real" encoder numbers?
      // Let's try actual incremental encoder numbers...
      rbytes += sprintf(replyBuffer, "OK,%.6f,%.6f,%.6f,%.6f\n",
#if 0
                        sh->myFixedPos.az,
                        sh->myFixedPos.el
#endif
                        sh->incr.az,
                        sh->incr.el,
                        sh->myTarget.az,  // calc azel /wo offsets
                        sh->myTarget.el
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    if (number_of_words < 3) {
      // 2 words, 1 parameter, only az given, needs el, too,
      // report "missing value"
      sureWrite(filedes, "MV,el\n", 6);
      goto lExit;
    }

    // At least 3 words, 2 parameters: azoff, eloff.
    if (getDHMS(words[1], /*deflt=>*/ eDegrees, &(sh->myFixedPos.az))) {
      sureWrite(filedes, "IV,az\n", 6);
      goto lExit;
    }
    if (getDHMS(words[2], /*deflt=>*/ eDegrees, &(sh->myFixedPos.el))) {
      sureWrite(filedes, "IV,el\n", 6);
      goto lExit;
    }

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    // Signal 'vsibcon' by incrementing the update flag counter.
    sh->vsibcliMode = ACM_azel;
    sh->vsibcliUpdates++;
    goto lExit;
  }  // if "azel"

  // xxx: "terminate" (for compatibility with the FS).

  if (strcmp(words[0], "GET")==0) {
    // xxx: Reply with real HTTP message. */
  }  // if "GET"

  if (strcmp(words[0], "reload")==0) {
    // xxx: Reload channel name & settings file.
    // Run configuration checks;
    // if no responses, generate error message
    // with 'sl.whatPathname()' telling the serial port used.
  }  // if "reload"

  // Nothing matched, let the inherited method do the work.
  retCode = cMessageProcessor::process_message_buffer(filedes,
                                                      buffer,
                                                      nbytes);
  goto lExit;

lExit:
  if (running) {
    free(running);
  }
  return retCode;
}  // process_message_buffer

void
cMPVsib::doBackground(void) {
#if 0
  time_t now;
  char replyBuffer[256];
  int i = this->nextChannel;

  now = time(NULL);
  if ( (now - ct[i].lastLogged) > ct[i].interval ) {
    // Log it.
    printf("%s\t%s\t%+g\n", getUTC(NULL), ct[i].name1, value);
    
    // Make sure the line gets immediately to stdout
    // (where 'arranger.pl' takes it and writes it into a file).
    fflush(stdout);

    ct[i].lastLogged = now;  // successful or not, for now...
  }  // if expired

  if (i < (ctSize-1)) {
    this->nextChannel++;
  } else {
    this->nextChannel = 0;
  }
#endif

  // Do default background processing.
  cMessageProcessor::doBackground();
}  // doBackground



// ----------------------------------------------------------------------
int
main (int argc, char **argv)
{
  cSocketServer ss;
  cMPVsib mp;

  int option_char;

  // Startup version message.
  fprintf (stderr, MY_NAME "$Id$\n");

  // Configuration file / command line option defaults.
  // Command line wins, default conf file is next,
  // and these hard-coded values are the last resort.
  int my_port_number = 5555;

  // Process default conf file.
  // xxx process_conf(/*file=>*/ "/etc/dammer.conf" );

  // Invokes member function `int operator ()(void);'
  while ((option_char = getopt (argc, argv, "c:p:")) != EOF)
    switch (option_char) {
       case 'c': {
         assert( optarg != NULL );
	 // xxx process_conf(/*file=>*/getopt.optarg);
	 break;
       }
       case 'p': {
	 my_port_number = atoi (optarg);
	 break;
       }
       case ':':  // missing parameter after a option letter
       case '?': {
	 fprintf (stderr,
		  "usage: %s [-p<port>]\n", argv[0]);
	 exit(EXIT_FAILURE);
       }
    }  // switch

  // Shared memory unique identifier (key) and the returned reference id.
  key_t shKey;
  int shId = -1;

  // Find and attach to 'vsibcon--vsibcli--xvsibcon' shared memory.
  shKey = fourCharLong('v','s','i','b');
  // Don't create, must exist (created by 'vsibcon')
  shId = shmget(shKey, sizeof(tSh), 0777);
  if (shId == -1) {
    perror(MY_NAME "shmget(): could not find 'vsibcon' shared memory");
    exit(EXIT_FAILURE);
  }
  assert( (sh = (ptSh)shmat(shId, NULL, 0)) != (ptSh)-1 );
  // Initialize 'vsibcli' portion of shared memory.
  // xxx: init like sh->target.az = sh->target.el = 0.0;

  // Prime the message processor.
  mp.init();

  // ------------------------------------------------------------
  // Start to listen and serve socket/port forever until killed.
  // Command lines from network are interpreted
  // and the values are written to shared memory for 'vsibcon'.
  ss.serve_socket (my_port_number, mp);
  // ------------------------------------------------------------

  // xxx: Is the following cleanup code ever executed?
  mp.cleanup();

#if 0
  printf("shKey=0x%08X, shId=%d, sh=%08X\n",
         shKey, shId, (unsigned int)sh);
#endif
  if ((shId != -1) && (sh != (ptSh)-1) && (sh != NULL)) {
    // It is not our business to remove the seqment,
    // 'vsibcon' takes care of managing removal.
    // assert( shmctl(shId, IPC_RMID, NULL) == 0 );
    assert( shmdt(sh) == 0 );
  }  // if shared memory was attached to

  return (EXIT_SUCCESS);
}  // main
